#include <iostream>
#include <assert.h>
#include "console.h"
#include "simpio.h"  // for getLine
#include "vector.h"  // for Vector
#include "lexicon.h"
#include "strlib.h"
using namespace std;

static string expand(const string& set) {
    Set<char> charset;
    size_t i = 1;
    while (i < set.size() - 1) {
        char low = set[i];
        char high = set[i + 1] == '-' ? set[i + 2] : low;
        i += set[i + 1] == '-' ? 3 : 1;
        for (char ch = low; ch <= high; ch++) charset += ch;
    }

    string expanded;
    for (char ch: charset) expanded += ch;
    return expanded;
}

static void split(const string& regex, string& first, bool& starred, string& rest) {
    size_t pos = !isalpha(regex[0]) ? regex.find(']') + 1 : 1;
    first = regex.substr(0, pos);
    starred = pos < regex.size() && regex[pos] == '*';
    if (starred) pos++;
    rest = regex.substr(pos);
}

static bool matches(const string& regex, const string& word) {
    if (regex.empty()) return word.empty(); // no regex? only accept empty string

    bool starred;
    string first, rest;
    split(regex, first, starred, rest);

    string set = first.size() == 1 ? first : expand(first);
    if (word.empty() || set.find(word[0]) == string::npos) {
        // leading component must be starred, rest must match everything
        return starred && matches(rest, word);
    }

    // consider first character being consumed by first
    if (matches(rest, word.substr(1))) return true;
    // if above call didn't work out and first wasn't starred, we're sad
    if (!starred) return false;

    // first item was starred, so first could be skipped entirely, provided rest matches
    // the full word.  Or the first character can match the starred first and the rest of
    // the word matches the full regex
    return matches(rest, word) || matches(regex, word.substr(1));
}

static void exerciseExpand(const string& str) {
    cout << "\"" << str << "\" -> \"" << expand(str) << "\"" << endl;
}

static void exerciseExpand() {
    exerciseExpand("[a-g]");
    exerciseExpand("[x-ya-g]");
    exerciseExpand("[a-empw-z]");
    exerciseExpand("[aeiou]");
    exerciseExpand("[a-ea-ed-fa-eeeee]");
}

static void exerciseSplit(const string& pattern) {
    string former, latter;
    bool repeat;
    split(pattern, former, repeat, latter);
    cout << "'" << pattern << "\" -> \"" << former << " (repeat? " << (repeat ? "yes" : "no") << "), \""
                                       << latter << "\"" << endl;
}

static void exerciseSplit() {
    exerciseSplit("awxyz");
    exerciseSplit("a*wyxz");
    exerciseSplit("[a-e]wyxz");
    exerciseSplit("[a-e]*wxyz");
    exerciseSplit("r");
    exerciseSplit("[a-z]*a[a-z]*e[a-z]*i[a-z]*o[a-z]*u[a-z]*y[a-z]*");
}

static void listAllMatches(const Lexicon& english, const string& regex) {
    for (const string& word: english) {
        if (matches(regex, word))
            cout << word << " matches " << regex << "." << endl;
    }
}

static void testMatches() {
    exerciseExpand();
    exerciseSplit();
    Lexicon english("dictionary.txt");
    while (true) {
        string regex = trim(getLine("Enter a regex: "));
        if (regex.empty()) break;
        listAllMatches(english, regex);
    }
}

int main() {
    testMatches();
    return 0;
}
